import traverse from "@babel/traverse";
import * as t from '@babel/types';
import {
    defaultProps,
    getConfigOptions,
    getForwardRefs,
    lifecycleHooks,
    manageCursorArrowFunctionExpressionVariableDeclaration,
    manageCursorFunctionDeclaration,
    processForwardRefs,
    skipSvgElements,
    stateAssignments,
    stateFunctions,
    stubsToReact,
    updateRefFunctionDeclarations,
    updateRefsIdentifier,
    useRefVariableDeclarator,
    useStateVariableDeclarator,
    useStoreVariableDeclarator
} from './transformers';
import { throwError } from '../utils';
import { jsxWithParens } from "./transformers/jsx";
import { ParserScope } from "../scope";
import { parseLoops } from "./transformers/loops";
import { parseTertiary } from "./transformers/tertiary";

export const parse = (ast: any, clean: boolean = false) => {
    const result = clean ? JSON.parse(JSON.stringify(ast)) : ast;

    const scope = new ParserScope();

    // validations and loading state definitions
    traverse(result, {
        ImportDeclaration(path) {
            if (path.node.source.value === 'react') {
                throwError('cannot import react!')
            }
        },
        FunctionDeclaration(path) {
            // update cursor
            manageCursorFunctionDeclaration(path, scope)
        },
        VariableDeclaration(path) {
            // update cursor
            manageCursorArrowFunctionExpressionVariableDeclaration(path, scope);
        },
        JSXElement(path) {
            // skip first
            skipSvgElements(path);

            // get loopy
            // TODO need to support recursion
            // e.g. For in For in Show in For in Show
            // e.g. we need to return true if we find a For/Show and while loop sub traverse 
            parseLoops(path, scope);
        },
        JSXAttribute(path) {
            getForwardRefs(path, scope);
        }
    });

    // teriary
    traverse(result, {
        FunctionDeclaration(path) {
            // update cursor
            manageCursorFunctionDeclaration(path, scope)
        },
        VariableDeclaration(path) {
            // update cursor
            manageCursorArrowFunctionExpressionVariableDeclaration(path, scope);
        },
        JSXElement(path) {
            // skip first
            skipSvgElements(path);

            // get loopy
            // TODO need to support recursion
            // e.g. For in For in Show in For in Show
            // e.g. we need to return true if we find a For/Show and while loop sub traverse 
            parseTertiary(path, scope);
        }
    });

    // get options
    traverse(result, {
        ExpressionStatement(path) {
            getConfigOptions(path, scope);
        },
        JSXElement(path) {
            skipSvgElements(path);
        }
    });

    // now traverse :) 
    traverse(result, {
        ImportDeclaration(path) {
            // useRef, useStore, useState, useEffect, userDefaultProps
            stubsToReact(path);
        },
        FunctionDeclaration(path) {
            // update cursor
            manageCursorFunctionDeclaration(path, scope);
        },
        VariableDeclaration(path) {
            // // update cursor
            manageCursorArrowFunctionExpressionVariableDeclaration(path, scope);
        },
        VariableDeclarator(path) {
            // useRef transformer
            useRefVariableDeclarator(path, scope);

            // useState transformer
            useStateVariableDeclarator(path, scope);

            // useStore transformer
            useStoreVariableDeclarator(path, scope);
        },
        AssignmentExpression(path) {
            // STATE
            // FROM state.property = value;
            // TO   setProperty(value);
            stateAssignments(path, scope);
        },
        CallExpression(path) {
            // state.hide(el);
            stateFunctions(path, scope);

            // useDefault props transformer
            defaultProps(path, scope);
        },
        ExpressionStatement(path) {
            // onMount, onUnMount, onUpdate
            lifecycleHooks(path);
        },

        JSXElement(path) {
            // always ensure to use parents when returning JSX
            jsxWithParens(path);
            skipSvgElements(path);
        },
        // TODO set configOptions for strings
        StringLiteral(path) {
            // @ts-ignore
            if (path.node.extra && path.node.extra.raw.startsWith('"')) {
                path.node.extra.raw = `'${path.node.value.replace(/'/g, "\\'")}'`;
            }
        }
    });

    // update all the state!
    // update all the refs!
    // move default props!

    traverse(result, {
        MemberExpression(path) {
            if (path.node.object.type === 'Identifier' &&
                path.node.object.name === 'state'
            ) {
                path.replaceWith(
                    path.node.property
                )
            }
        },
        JSXElement(path) {
            skipSvgElements(path);
        }
    })

    traverse(result, {
        FunctionDeclaration(path) {
            // update cursor
            manageCursorFunctionDeclaration(path, scope);
        },
        VariableDeclaration(path) {
            // update cursor
            manageCursorArrowFunctionExpressionVariableDeclaration(path, scope);
        },
        Identifier(path) {
            updateRefsIdentifier(path, scope);
        },
        JSXElement(path) {
            skipSvgElements(path);
        },
        JSXAttribute(path) {
            processForwardRefs(path, scope);
        }
    });

    // top-level forward refs
    traverse(result, {
        FunctionDeclaration(path) {
            // update cursor
            manageCursorFunctionDeclaration(path, scope);
            // update ref methods
            updateRefFunctionDeclarations(path, scope);
        },
        VariableDeclaration(path) {
            // update cursor
            manageCursorArrowFunctionExpressionVariableDeclaration(path, scope);
        },
        JSXElement(path) {
            skipSvgElements(path);
        }
    });


    if (scope.configOptions?.rsc?.componentType === 'client') {
        result.program.directives = [
            t.directive(
                t.directiveLiteral(
                    'use client'
                )
            )
        ];
    }

    const imports = [];
    const exports = [];

    // TODO useMetadata options to set
    imports.push(
        t.importDeclaration(
            [t.importNamespaceSpecifier(
                t.identifier('React')
            )],
            t.stringLiteral('react')
        )
    );

    // add forwardRef imports
    if (scope.components.some(
        component => component.hasForwardRef()
    )) {
        imports.push(
            t.importDeclaration(
                [
                    t.importSpecifier(
                        t.identifier('forwardRef'),
                        t.identifier('forwardRef')
                    )
                ],
                t.stringLiteral('react')
            )
        );
        if (scope.defaultForwardRefExport) {
            exports.push(
                t.exportDefaultDeclaration(
                    t.identifier(scope.defaultForwardRefExport)
                )
            )
        }
    }
    result.program.body = [
        ...imports,
        ...result.program.body,
        ...exports
    ];

    return {
        scope,
        ast: result
    };
};
