import * as t from '@babel/types';
import traverse, { NodePath } from '@babel/traverse';
import { throwError } from '../../utils';
import { ParserScope } from '../../scope';

export const isJSXReturn = (path: NodePath<t.ReturnStatement>) => {
    if (path.node.argument &&
        (
            path.node.argument.type === 'JSXElement' ||
            path.node.argument.type === 'JSXFragment'
        )) {
        return true;
    }
    return false;
};

// function MyComponentFunc() { };
// export function myFunction2() { };
// export default function myFunction3() { }

export const isTopLevelDefinition = (
    path: NodePath<t.FunctionDeclaration | t.VariableDeclaration>
) => {
    let atTopLevel = false;
    if (
        path.parent.type === 'ExportDefaultDeclaration' ||
        path.parent.type === 'ExportNamedDeclaration'
    ) {
        if (path.parentPath.parent.type === 'Program') {
            atTopLevel = true;
        }
    } else if (path.parent.type === 'Program') {
        atTopLevel = true;
    }
    return atTopLevel;
};

export const isTopLevelFunctionDeclaration = (
    path: NodePath<t.FunctionDeclaration>
) => {
    let hasJSXReturn = false;

    if (!isTopLevelDefinition(path)) return;

    // Create a Program node that wraps the function body
    const programNode = t.file(
        t.program([
            t.expressionStatement(
                t.functionExpression(null, path.node.params, path.node.body)
            )
        ])
    );

    // Traverse the cloned body
    traverse(programNode, {
        ReturnStatement(returnPath) {
            if (isJSXReturn(returnPath)) {
                hasJSXReturn = true;
                returnPath.stop();
            }
        }
    });

    if (hasJSXReturn) {
        if (!path.node.id && !path.node.id.name) {
            return throwError('cannot define component function with no name');
        }

        return true;
    }
    return false;
}

export const manageCursorFunctionDeclaration = (
    path: NodePath<t.FunctionDeclaration>,
    scope: ParserScope
) => {
    if (!isTopLevelFunctionDeclaration(path)) return;
    scope.setCursor(path.node.id.name);
    scope.addComponent(path.node.id.name);
}

// export const MyComponentArrowFunc = () => { };
// const MyComponentArrowFunc = () => { };
export const isTopLevelArrowFunctionExpressionVariableDeclaration = (
    path: NodePath<t.VariableDeclaration>
) => {
    let hasJSXReturn = false;

    if (!isTopLevelDefinition(path)) return;

    // is there an ArrowFunctionExpression ?
    let isArrowFunctionExpression = false;
    const decl: t.VariableDeclarator = path.node.declarations[0];
    if (decl.type === 'VariableDeclarator') {
        if (decl.init.type === 'ArrowFunctionExpression') {
            isArrowFunctionExpression = true;
        }
    }

    if (!isArrowFunctionExpression) return;

    const functionExpression: t.ArrowFunctionExpression = decl.init as t.ArrowFunctionExpression;
    // Create a Program node that wraps the function body
    const programNode = t.file(
        t.program([
            functionExpression.body as any
        ])
    );

    // Traverse the cloned body
    traverse(programNode, {
        ReturnStatement(returnPath) {
            if (isJSXReturn(returnPath)) {
                hasJSXReturn = true;
                returnPath.stop();
            }
        }
    });

    if (hasJSXReturn) {
        if (decl.id.type !== 'Identifier') {
            return throwError('arrow function without Identifier name at top level');
        }
        return true;
    }
    return false;
}

export const manageCursorArrowFunctionExpressionVariableDeclaration = (
    path: NodePath<t.VariableDeclaration>,
    scope: ParserScope
) => {
    const decl: t.VariableDeclarator = path.node.declarations[0];
    if (!isTopLevelArrowFunctionExpressionVariableDeclaration(path)) return;
    const id: t.Identifier = decl.id as t.Identifier;
    scope.setCursor(id.name);
    scope.addComponent(id.name);
}


