import * as t from '@babel/types';
import { NodePath } from '@babel/traverse';
import { throwError } from '../../utils';
import { ParserScope } from '../../scope';
import { isTopLevelFunctionDeclaration } from './component';

export const updateRefsIdentifier = (
    path: NodePath<t.Identifier>,
    scope: ParserScope
) => {
    // in a MemberExpression?
    if (path.parent.type === 'MemberExpression') {
        return;
    }

    // in a VariableDeclarator?
    if (path.parent.type === 'VariableDeclarator') {
        return;
    }

    // is a ref there?
    if (!scope.currentComponent()?.hasRef(path.node.name)) {
        return;
    }

    // if a ref passing a ref to a component, leave it
    if (path.parent.type == 'JSXExpressionContainer') {
        if (path.parentPath.parent.type === 'JSXAttribute') {
            const attr: t.JSXAttribute = path.parentPath.parent;
            if (attr.name.name === 'ref') {
                return;
            }
        }
    }

    // if a forwardRef passing a ref to a component, leave it
    if (path.parent.type == 'LogicalExpression') {
        if (path.parentPath.parent.type == 'JSXExpressionContainer') {
            if (path.parentPath.parentPath.parent.type === 'JSXAttribute') {
                const attr: t.JSXAttribute = path.parentPath.parentPath.parent;
                if (attr.name.name === 'ref') {
                    return;
                }
            }
        }
    }

    path.replaceWith(
        t.memberExpression(
            t.identifier(path.node.name),
            t.identifier('current')
        )
    );
};

export const useRefVariableDeclarator = (
    path: NodePath<t.VariableDeclarator>,
    scope: ParserScope
) => {

    if (path.node.init &&
        path.node.init.type === 'CallExpression' &&
        path.node.init.callee &&
        path.node.init.callee.type === 'Identifier' &&
        path.node.init.callee.name === 'useRef') {

        const node = path.node;
        if (node.id.type !== 'Identifier') {
            return throwError('useRef must use an Identifier');
        }

        if (scope.currentComponent().hasRef(node.id.name)) {
            // throwError('refs variable is already defined!')
        } else {
            scope.currentComponent().addRef({
                name: node.id.name
            })
        }
    }
};

export const isForwardRef = (
    path: NodePath<t.JSXAttribute>
) => {
    if (path.node.name.type === 'JSXIdentifier') {
        if (path.node.name.name !== 'ref') return;
        if (path.node.value.type === 'JSXExpressionContainer') {
            if (path.node.value.expression.type === 'MemberExpression') {
                if (path.node.value.expression.object.type === 'Identifier') {
                    return true;
                }
            }
        }
    }
    return false;
};

export const getForwardRefs = (
    path: NodePath<t.JSXAttribute>,
    scope: ParserScope
) => {
    if (!isForwardRef(path)) return;

    const expr: t.JSXExpressionContainer = path.node.value as t.JSXExpressionContainer;
    const memb: t.MemberExpression = expr.expression as t.MemberExpression;
    const iden: t.Identifier = memb.object as t.Identifier;
    const prop: t.Identifier = memb.property as t.Identifier;

    // TODO
    // TODO
    // TODO
    // TODO get the current component props name as a variable
    // THIS means you can have MyComponent(myProps) vs Component(props)
    if (iden.name !== 'props') return;
    if (prop.type !== 'Identifier') return;

    const name = prop.name;
    if (scope.currentComponent().hasRef(name)) return;

    scope.currentComponent().addRef({
        name,
        forward: true
    })
};

export const processForwardRefs = (
    path: NodePath<t.JSXAttribute>,
    scope: ParserScope
) => {
    if (!isForwardRef(path)) return;

    const expr: t.JSXExpressionContainer = path.node.value as t.JSXExpressionContainer;
    const memb: t.MemberExpression = expr.expression as t.MemberExpression;
    const iden: t.Identifier = memb.object as t.Identifier;
    const prop: t.Identifier = memb.property as t.Identifier;



    // TODO
    // TODO
    // TODO
    // TODO get the current component props name as a variable
    // THIS means you can have MyComponent(myProps) vs Component(props)
    if (iden.name !== 'props') return;
    if (prop.type !== 'Identifier') return;
    const name = prop.name;
    if (!scope.currentComponent().hasRef(name)) return;

    path.replaceWith(
        t.jsxAttribute(
            t.jsxIdentifier('ref'),
            t.jsxExpressionContainer(
                t.logicalExpression(
                    '??',
                    memb,
                    prop // COPY property
                )
            )
        )
    );

};

export const updateRefFunctionDeclarations = (
    path: NodePath<t.FunctionDeclaration>,
    scope: ParserScope
) => {
    if (!isTopLevelFunctionDeclaration(path)) return;
    const current = scope.currentComponent(true);
    if (!current.hasForwardRef()) return;
    const ref = current.findForwardRef();

    if (path.parentPath.node.type === 'ExportDefaultDeclaration') {
        if (path.node.id.type !== 'Identifier') return;

        scope.defaultForwardRefExport = path.node.id.name;

        const params = path.node.params;
        if (params.length !== 1) {
            throwError('forwardRefs require one property, props!');
        }
        if (params[0].type !== 'Identifier') {
            throwError('forwardRefs require one Identifier, props!');
        }
        const id: t.Identifier = params[0] as t.Identifier;

        let typeRef;
        if (id.typeAnnotation && id.typeAnnotation.type === 'TSTypeAnnotation') {
            typeRef = id.typeAnnotation.typeAnnotation;
        }

        const refIdent = t.identifier(ref.name);
        refIdent.typeAnnotation = t.tsTypeAnnotation(
            t.tsIndexedAccessType(
                typeRef,
                t.tsLiteralType(
                    t.stringLiteral(ref.name)
                )
            )
        )

        const callExpr = t.callExpression(
            t.identifier('forwardRef'),
            [
                t.functionExpression(
                    t.identifier(path.node.id.name),
                    [
                        ...path.node.params,
                        refIdent

                    ],
                    path.node.body,
                    path.node.generator,
                    path.node.async
                )
            ]
        );

        callExpr.typeParameters = t.tsTypeParameterInstantiation(
            [
                t.tsAnyKeyword(),
                typeRef
            ]
        )
        path.parentPath.replaceWith(
            t.variableDeclaration(
                'const',
                [
                    t.variableDeclarator(
                        t.identifier(path.node.id.name),
                        callExpr
                    )
                ]
            )
        )
    }
};