import generate from '@babel/generator';
import * as t from '@babel/types';
import { NodePath } from '@babel/traverse';
import { makeSetState } from '../utils';
import { throwError } from '../../utils';
import { ParserScope } from '../../scope';

export const validateStateVariableDeclarator = (
    path: NodePath<t.VariableDeclarator>,
    scope: ParserScope
) => {
    if (path.node.init &&
        path.node.init.type === 'CallExpression' &&
        path.node.init.callee &&
        path.node.init.callee.type === 'Identifier' &&
        path.node.init.callee.name === 'useState') {

        const node = path.node;
        if (node.id.type !== 'ArrayPattern') {
            return throwError('useState must use an array');
        }
        if (node.id.elements.length !== 2) {
            return throwError('useState must use an array of length 2');
        }

        if (node.id.elements[0].type !== 'Identifier') {
            return throwError('useState first element must be Identifier');
        }

        if (node.id.elements[1].type !== 'Identifier') {
            return throwError('useState second element must be Identifier');
        }

        const current = scope.currentComponent();
        if (!current) {
            console.log(generate(node).code)
            throwError('cursor missing')
        }

        // store in hash
        const name = node.id.elements[0].name;
        if (!scope.currentComponent()?.state?.exists(name)) {

        } else {
            scope.currentComponent().addProp({
                name,
                derived: 'useState',
                from: 'ObjectProperty',
            });
        }
    }
};

export const useStateVariableDeclarator = (
    path: NodePath<t.VariableDeclarator>,
    scope: ParserScope
) => {

    validateStateVariableDeclarator(path, scope);

    if (path.node.init &&
        path.node.init.type === 'CallExpression' &&
        path.node.init.callee &&
        path.node.init.callee.type === 'Identifier' &&
        path.node.init.callee.name === 'useState') {


        // process node
        if (path.node.init.arguments[0].type === 'ArrowFunctionExpression') {
            return;
        }

        // update arguments
        path.node.init.arguments = [
            t.arrowFunctionExpression(
                [],
                path.node.init.arguments[0] as any,
                false
            )
        ];
    }
};

export const useStoreVariableDeclarator = (
    path: NodePath<t.VariableDeclarator>,
    scope: ParserScope
) => {
    if (path.node.init &&
        path.node.init.type === 'CallExpression' &&
        path.node.init.callee &&
        path.node.init.callee.type === 'Identifier' &&
        path.node.init.callee.name === 'useStore') {

        const args: t.ObjectExpression[] = path.node.init.arguments as t.ObjectExpression[];
        const objectArgument: t.ObjectExpression = args[0];
        if (t.isObjectExpression(objectArgument) && objectArgument.properties.length > 0) {
            const newDeclarations = objectArgument.properties.map(property => {

                switch (property.type) {
                    case 'ObjectMethod': {
                        const method: t.ObjectMethod = property;
                        const key: t.Identifier = method.key as t.Identifier;

                        if (key.type !== 'Identifier') {
                            throwError('store methods must be Identifiers')
                        }

                        if (scope.currentComponent().findProp(key.name)) {
                            throwError('state variable is already defined!')
                        }

                        scope.currentComponent().addProp({
                            name: key.name,
                            derived: 'useStore',
                            from: 'ObjectMethod'
                        });

                        return t.functionDeclaration(
                            t.identifier(key.name),
                            method.params,
                            method.body,
                            method.generator,
                            method.async
                        )


                    }
                    case 'ObjectProperty': {
                        const prop: t.ObjectProperty = property;
                        const key: t.Identifier = prop.key as t.Identifier;

                        if (key.type !== 'Identifier') {
                            throwError('store object properties must be Identifier')
                        }

                        if (!scope.currentComponent()?.findProp(key.name)) {
                            // throwError('state variable is already defined!');
                        }

                        scope.currentComponent().addProp({
                            name: key.name,
                            derived: 'useStore',
                            from: 'ObjectProperty'
                        });

                        const stateName = t.identifier(key.name);
                        const stateSetterName = t.identifier(makeSetState(key.name));

                        return t.variableDeclaration('const', [
                            t.variableDeclarator(
                                t.arrayPattern([stateName, stateSetterName]),
                                t.callExpression(t.identifier('useState'), [t.arrowFunctionExpression([], prop.value as any)])
                            )
                        ]);
                    }
                    default:
                        throwError('store must be ObjectMethod or ObjectProperty')

                }

                // @ts-ignore
                const prop: t.ObjectProperty = property;
                const key: t.Identifier = prop.key as t.Identifier;

                if (key.type !== 'Identifier') {
                    throwError('store object properties must be Identifier')
                }
                const stateName = t.identifier(key.name);
                const stateSetterName = t.identifier(makeSetState(key.name));

                return t.variableDeclaration('const', [
                    t.variableDeclarator(
                        t.arrayPattern([stateName, stateSetterName]),
                        t.callExpression(t.identifier('useState'), [t.arrowFunctionExpression([], prop.value as any)])
                    )
                ]);
            });

            // this updates the array above! awesome!
            path.parentPath.replaceWithMultiple(newDeclarations);
        }
    }

};

export const stateAssignments = (
    path: NodePath<t.AssignmentExpression>,
    scope: ParserScope
) => {
    // STATE
    // FROM state.property = value;
    // TO   setProperty(value);
    if (path.node.type === 'AssignmentExpression') {
        if (path.node.left.type === 'MemberExpression') {
            if (path.node.left.object.type === 'Identifier') {
                if (path.node.left.object.name === 'state') {
                    // left: path.node.left.property as any
                    if (path.node.left.property.type !== 'Identifier') {
                        return throwError('set state property not yet supported.');
                    }

                    // set that state variable is used!
                    const name = path.node.left.property.name;
                    const current = scope.currentComponent(true);
                    if (current.findProp(name)) {
                        current.findProp(name).setUsed(true);
                    } else {
                        throwError(scope.errorMsg('PROP_SET_NOT_FOUND', { prop: name }));
                    }

                    path.replaceWith(t.expressionStatement(
                        t.callExpression(
                            t.identifier(
                                makeSetState(path.node.left.property.name)
                            ),
                            [path.node.right as any]
                        )
                    ));
                }
            }
        }
    }
};


export const stateFunctions = (
    path: NodePath<t.CallExpression>,
    scope: ParserScope
) => {
    // STATE
    // FROM state.method(value)
    // TO   method(value);
    if (path.node.type === 'CallExpression') {
        if (path.node.callee.type === 'MemberExpression') {
            if (path.node.callee.object.type === 'Identifier') {
                if (path.node.callee.object.name === 'state') {
                    // left: path.node.callee.property as any
                    if (path.node.callee.property.type !== 'Identifier') {
                        return throwError('set state property not yet supported.');
                    }

                    // set that state variable is used!
                    const name = path.node.callee.property.name;
                    const current = scope.currentComponent();
                    if (!current) {
                        throwError(scope.errorMsg('PROP_FUNC_COMP_NOT_FOUND', { prop: name }));
                    }
                    if (current.state.exists(name)) {
                        current.state.get(name).setUsed(true);
                    } else {
                        throwError(scope.errorMsg('PROP_FUNC_NOT_FOUND', { prop: name }));
                    }

                    path.replaceWith(t.callExpression(
                        path.node.callee.property as any,
                        path.node.arguments as any
                    ));
                }
            }
        }
    }
};

export const defaultProps = (
    path: NodePath<t.CallExpression>,
    scope: ParserScope
) => {
    if (path.node.callee.type === 'Identifier' &&
        path.node.callee.name === 'useDefaultProps') {

        const args = path.node.arguments;

        const Component: t.Identifier = args[0] as t.Identifier;
        const Props: t.ObjectExpression = args[1] as t.ObjectExpression;

        if (Component.type !== 'Identifier') {
            throwError('useDefaultProps first argument must be an Identifier');
        }
        if (Props.type !== 'ObjectExpression') {
            throwError('useDefaultProps first argument must be an ObjectExpression');
        }

        // replace with Component.defaultProps = Props;
        path.replaceWith(t.expressionStatement(
            t.assignmentExpression('=',
                t.memberExpression(
                    t.identifier(Component.name),
                    t.identifier('defaultProps'),
                    false
                ),
                Props
            )
        ));

    }
};